#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for LOCKKEY configuration."""

import os
import libpal.palcmd as palcmd
from libpal.util import RemountRoot
from iobase.cmdexec import (CmdExec, NULL, PIPE)
from iobase.netut import is_inet6_addr
from iobase.exception import (ExecError, InvalidArgError)
from pal_info import (ReqTypes, Results)
from pal_config import (SYSFS_DIR, DEV_DIR, CRYPT_LOCKKEY_DIR, GNUPG_DIR)
from pal_config import (ENCKEY_CMD, GPG_CMD)

### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#

VAR = "%s"
# enckey --get --usb [dev param path] --path [key path] --oprintable
GET_KEY_ARG_FORMAT = \
    [ENCKEY_CMD, "--get",  "--usb", VAR, "--path", VAR, "--oprintable"]
GET_KEY_VAR_INDEX_LIST = [3, 5]

# enckey --set --usb [dev param path]  --path [key path] --iprintable
WRITE_KEY_ARG_FORMAT = \
    [ENCKEY_CMD, "--set", "--usb", VAR, "--path", VAR, "--iprintable"]
WRITE_KEY_VAR_INDEX_LIST = [3, 5]

# enckey --create --key [encode key] --usb [dev param path] --path [key path]
CREATE_KEY_ARG_FORMAT = \
    [ENCKEY_CMD, "--create", "--key", VAR, "--usb", VAR, "--path", VAR]
CREATE_KEY_VAR_INDEX_LIST = [3, 5, 7]

# enckey --get --key [encode key] --usb [dev param path] --path [key path]
EXTRACT_FROM_USB_ARG_FORMAT = \
    [ENCKEY_CMD, "--get", "--key", VAR, "--usb", VAR, "--path", VAR]
EXTRACT_FROM_USB_VAR_INDEX_LIST = [3, 5, 7]

# enckey --get --key [encode key] --keyserver [host:port] --macaddr [mac]
EXTRACT_FROM_KEYSERVER_ARG_FORMAT = \
    [ENCKEY_CMD, "--get", "--key", VAR, "--keyserver", VAR, "--macaddr", VAR]
EXTRACT_FROM_KEYSERVER_VAR_INDEX_LIST = [3, 5, 7]

# enckey --keyserver [hostaddr:port]  --check-connect
CHK_KEYSRV_CONNECTION_ARG_FORMAT = \
    [ENCKEY_CMD, "--keyserver", VAR, "--check-connect"]
CHK_KEYSRV_CONNECTION_VAR_INDEX_LIST = [2]

## X path ###
KEYSERVER_XPATH = "./keyserverlist/keyserver"
USBDEV_NAME_XPATH = "./device/name[@type='usb']"
ENCODE_FILE_PATH_XPATH = "./path[@type='encode']"
DECODE_FILE_PATH_XPATH = "./path[@type='decode']"

### XML Tag ###
MAC_NAME = "mac"
LOCKKEY_NAME = "lockkey"
KEY_NAME = "key"
KEYSERVER_NAME = "keyserver"
HOSTNAME_NAME = "hostname"
PORT_NAME = "port"
OPTION_NAME = "option"

### XML Attribute ###
CREATE_ATTR_VAL = "create"
EXTRACT_ATTR_VAL = "extract"
WRITE_ATTR_VAL = "write"
KEYSERVER_ATTR = "keyserver"

### Request option ###
OPERATION_OPT = "operation"

### Some Message ###
INVALID_ROOT_TAG_MSG = "Invalid tag name !!"
FAIL_EXECUTE_CMD = "Fail to execute command !!"


class PalLockkey(palcmd.PalCmd):
    """PAL Lockkey command implementation class."""

    class __CreateError(Exception):
        "Create Error"
        pass

    class __ExtractError(Exception):
        "Extract Error"
        pass

    class __WriteError(Exception):
        "Write Error"
        pass

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != LOCKKEY_NAME:
            raise InvalidArgError(INVALID_ROOT_TAG_MSG)

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_in, std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError(FAIL_EXECUTE_CMD)

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       stdin=std_in,
                       wait=is_wait,
                       raise_exc=is_enable_exception).execute()

    def __save_value(self, file_path, value):
        """Save specified value into specified file."""

        dir_path = os.path.dirname(file_path)
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

        with open(file_path, "w") as fh:
            fh.write(value)

    def __force_make_dirs(self, dir_path):
        """make specified directory."""
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    def __convert_real_devname(self, dev_name):
        """Convert synbol device name to real device name.
           If dev_name is already real this method just returns it.
        """

        dev_path = os.path.join(DEV_DIR, dev_name)
        return os.path.basename(os.path.realpath(dev_path))

    def __get_device_param_dir_path(self, dev_name):
        """Get device parameter dir path.
           The path is used by enckey command to encode by
           USB serial number.
            ex. "/sys/blok/sdg/../../../7:0:0/../../../../"
        """

        sys_blk_dev_path = os.path.join(SYSFS_DIR,
                                        "block",
                                        self.__convert_real_devname(dev_name))

        dev_tmp_dir = os.path.join(sys_blk_dev_path, "device")

        return os.path.join(sys_blk_dev_path,
                            os.path.realpath(dev_tmp_dir),
                            "../../../..")

    def __create_lockkey(self, entry):
        """Create lockkey."""

        mac = \
            entry.get().get_node(MAC_NAME).get_text().replace(":", "").upper()
        enc_file_path = entry.get().get_node(ENCODE_FILE_PATH_XPATH).get_text()
        self.__force_make_dirs(os.path.dirname(enc_file_path))
        usbdev_name = entry.get().get_node(USBDEV_NAME_XPATH).get_text()
        usbdev_param_path = self.__get_device_param_dir_path(usbdev_name)

        proc = self.__execute_cmd(False, False,
                                  NULL, PIPE, NULL,
                                  CREATE_KEY_ARG_FORMAT,
                                  CREATE_KEY_VAR_INDEX_LIST,
                                  [mac,
                                   usbdev_param_path,
                                   enc_file_path])
        if 0 != proc.wait():
            raise self.__CreateError

        crypt_path = os.path.join(CRYPT_LOCKKEY_DIR, mac + ".crypt")
        if os.path.exists(CRYPT_LOCKKEY_DIR):
            if os.path.exists(crypt_path):
                os.remove(crypt_path)
        else:
            os.mkdir(CRYPT_LOCKKEY_DIR)

        with RemountRoot():
            getcmd_list = [ENCKEY_CMD, "--get",  "--usb",
                    usbdev_param_path, "--path", enc_file_path]
            gpgcmd_list = [GPG_CMD, "--homedir", GNUPG_DIR,
                    "-aer", "LANDISK", "-o", crypt_path]
            getcmd = CmdExec(getcmd_list, stderr=NULL)
            gpgcmd = CmdExec(gpgcmd_list, stdout=NULL, stderr=NULL)
            proc = getcmd | gpgcmd
            if 0 != proc.wait():
                raise self.__CreateError

        return

    def __extract_lockkey_from_usb(self, entry):
        """Get decoded lockkey from specified usb memory.
           And save it to specified file path.
        """

        mac = \
            entry.get().get_node(MAC_NAME).get_text().replace(":", "").upper()
        enc_file_path = entry.get().get_node(ENCODE_FILE_PATH_XPATH).get_text()
        usbdev_name = entry.get().get_node(USBDEV_NAME_XPATH).get_text()
        usbdev_param_path = self.__get_device_param_dir_path(usbdev_name)

        proc = self.__execute_cmd(False, False,
                                  NULL, PIPE, NULL,
                                  EXTRACT_FROM_USB_ARG_FORMAT,
                                  EXTRACT_FROM_USB_VAR_INDEX_LIST,
                                  [mac, usbdev_param_path, enc_file_path])
        key_value = proc.stdout.read()
        if 0 != proc.wait():
            raise self.__ExtractError

        # save key into file
        dec_file_path = entry.get().get_node(DECODE_FILE_PATH_XPATH).get_text()
        self.__force_make_dirs(os.path.dirname(dec_file_path))
        self.__save_value(dec_file_path, key_value)
        return

    def __inet6_conv(self, hostname):

        if is_inet6_addr(hostname):
            return "[" + hostname + "]"
        return hostname

    def __extract_lockkey_from_keyserver(self, entry):
        """Get decoded lockkey from specified keyserver.
           And save it to specified file path.
        """

        mac = \
            entry.get().get_node(MAC_NAME).get_text().replace(":", "").upper()
        keyserver_node_list = entry.get().get_node_list(KEYSERVER_XPATH)

        host_port_list = []
        for keyserver_node in keyserver_node_list:
            hostname = keyserver_node.get_node(HOSTNAME_NAME).get_text()
            port = keyserver_node.get_node(PORT_NAME).get_text()

            hostname = self.__inet6_conv(hostname)

            host_port_list.append(hostname + ":" + port)

        # get locckey key value
        proc = self.__execute_cmd(False, False,
                                  NULL, PIPE, NULL,
                                  EXTRACT_FROM_KEYSERVER_ARG_FORMAT,
                                  EXTRACT_FROM_KEYSERVER_VAR_INDEX_LIST,
                                  [mac, ",".join(host_port_list), mac])
        key_value = proc.stdout.read()
        if 0 != proc.wait():
            raise self.__ExtractError

        # save key into file
        dec_file_path = entry.get().get_node(DECODE_FILE_PATH_XPATH).get_text()
        self.__force_make_dirs(os.path.dirname(dec_file_path))
        self.__save_value(dec_file_path, key_value)
        return

    def __extract_lockkey(self, entry):
        """Get decoded lockkey and save it to specified file path."""

        opt_node = entry.get().get_node(OPTION_NAME)
        is_keyserver = opt_node.is_set_to_true(KEYSERVER_ATTR)

        if is_keyserver:
            self.__extract_lockkey_from_keyserver(entry)

        else:
            self.__extract_lockkey_from_usb(entry)

        return

    def __write_lockkey(self, entry):
        """Save key to specified file path.
           The key is encoded before saving.
        """

        enc_file_path = entry.get().get_node(ENCODE_FILE_PATH_XPATH).get_text()
        encoded_key = entry.get().get_node(KEY_NAME).get_text()
        self.__force_make_dirs(os.path.dirname(enc_file_path))
        usbdev_name = entry.get().get_node(USBDEV_NAME_XPATH).get_text()
        usbdev_param_path = self.__get_device_param_dir_path(usbdev_name)

        proc = self.__execute_cmd(False, False,
                                  PIPE, PIPE, NULL,
                                  WRITE_KEY_ARG_FORMAT,
                                  WRITE_KEY_VAR_INDEX_LIST,
                                  [usbdev_param_path,
                                   enc_file_path])

        proc.stdin.write(encoded_key)
        proc.stdin.close()

        if 0 != proc.wait():
            raise self.__WriteError
        return

    def __query_handler(self, entry):
        """Handle query request."""

        self.__check_root_tag(entry)

        enc_file_path = entry.get().get_node(ENCODE_FILE_PATH_XPATH).get_text()
        usbdev_name = entry.get().get_node(USBDEV_NAME_XPATH).get_text()
        usbdev_param_path = self.__get_device_param_dir_path(usbdev_name)

        proc = self.__execute_cmd(False, False,
                                  NULL, PIPE, NULL,
                                  GET_KEY_ARG_FORMAT,
                                  GET_KEY_VAR_INDEX_LIST,
                                  [usbdev_param_path,
                                   enc_file_path])
        key = proc.stdout.read()

        entry.get().append_child(KEY_NAME, key)

        if 0 != proc.wait():
            entry.set_result(Results.QUERY_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __set_lockkey_handler(self, entry):
        """Handle set_lockkey request."""

        self.__check_root_tag(entry)

        ctrl_funcs = {CREATE_ATTR_VAL: self.__create_lockkey,
                      EXTRACT_ATTR_VAL: self.__extract_lockkey,
                      WRITE_ATTR_VAL: self.__write_lockkey}

        op = entry.get_options()[OPERATION_OPT]

        try:
            ctrl_funcs[op](entry)
            entry.set_result(Results.OK)
        except self.__CreateError:
            entry.set_result(Results.CREATE_ERROR)
        except self.__ExtractError:
            entry.set_result(Results.GET_ERROR)
        except self.__WriteError:
            entry.set_result(Results.WRITE_ERROR)
        except:
            raise

        return entry

    def __check_keyserver_handler(self, entry):
        """Handle check_keyserver request."""

        self.__check_root_tag(entry)

        keyserver_node = entry.get().get_node(KEYSERVER_NAME)
        hostname = keyserver_node.get_node(HOSTNAME_NAME).get_text()
        port = keyserver_node.get_node(PORT_NAME).get_text()

        hostname = self.__inet6_conv(hostname)

        proc = self.__execute_cmd(False, False,
                                  NULL, NULL, NULL,
                                  CHK_KEYSRV_CONNECTION_ARG_FORMAT,
                                  CHK_KEYSRV_CONNECTION_VAR_INDEX_LIST,
                                  [hostname + ":" + port])
        if 0 != proc.wait():
            entry.set_result(Results.CHECK_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.SET_LOCKKEY, self.__set_lockkey_handler)
        self.set_handler(ReqTypes.CHECK_KEYSERVER,
                         self.__check_keyserver_handler)

#
# Main
#

if __name__ == "__main__":
    PalLockkey(option=palcmd.PalCmdOption()).handle_request()
